#!/usr/bin/python3
from asyncore import compact_traceback
import ctypes
from multiprocessing.sharedctypes import Value
from select import select
import subprocess
from ctypes import *
from enum import Enum
import socket
import psutil
import json
import threading
import re
import time
import datetime
from pydbus import SystemBus
import dbus
import syslog
import os
import schedule
from gi.repository import GLib, Gtk, Gio
import gi
gi.require_version('Gtk', '3.0')

from kysecKeyProcessControl import *

class processInfo:
    def __init__(self, pid):
        self.pid = pid
        self.status = "run"
        self.command = ""
        self.apath = ""

        try:
            #commpath = "/proc/" + self.pid +"/comm"
            #with open(commpath) as fp:
            #    self.command = fp.readline().strip()
        
            absolutepath = "/proc/" + self.pid + "/exe"
            self.apath = os.readlink(absolutepath)
            self.command = self.apath[self.apath.rfind("/")+1:]
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "application procfs not exist.")
            self.command = ""
            self.apath = ""

class processMonitor:
    """
    <node>
        <interface name='com.kylin.processMonitor.startend'>
            <method name='setProcess' >
                <arg type='s' direction='in' name='list'/>
                <arg type='i' direction='out' name='ret'/>
            </method>
            <method name='getProcess' >
                <arg type='as' direction='out' name='list'/>
            </method>

            <method name='getCommandbyPid' >
                <arg type='s' direction='in' name='pid' />
                <arg type='s' direction='out' name='command' />
            </method>
        </interface>
    </node>
    """

    def __init__(self):
        syslog.syslog(syslog.LOG_DEBUG, "processMonitor init")
        try:
            self.log = SystemBus().get("org.log.sys_transmit", "/org/log/sys_transmit")
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "log service is error")
            thread_load_transmit = threading.Thread(target = self.load_transmit,args=())
            thread_load_transmit.setDaemon(True)
            thread_load_transmit.start()

        self.keyProcessListPath = "/etc/key-process-list"
        self.auditlogPath = "/var/log/audit/audit.log"
        #self.auditlogPath = "./test.log"
        self.processList = []  #保存关键进程信息，通过pid获取应用名称信息
        self.officeExecList = self.readProcessList() #保存配置的关键应用

        #开启
        logMonitorThread = threading.Thread(target=self.logMonitor, args=(self,))
        logMonitorThread.start()

        print("start processMonitor...")

        #定时任务
        schedule.every().day.at('00:00').do(self.dailySend)

    @property
    def get_host_name(self):
        """获取当前用户名"""
        try:
            bus2 = dbus.SystemBus()
            proxy = bus2.get_object("com.kylin.kydevmonit.hedronclient","/kydevmonit/hedronclient")
            username = proxy.get_current_user()
            syslog.syslog(syslog.LOG_DEBUG, "get_current_user {}".format(username))
            if(username == '' or username == ' '):
                username = '未登录'
        except:
            syslog.syslog(syslog.LOG_DEBUG, "call hedronclient get_current_user error")
            username = '未登录'
        return  username

    @property
    def get_host_time(self):
        """获取本机时间
        return: cur_time
        """
        cur_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        return cur_time

    @property
    def get_host_ip(self):
        """
        查询本机ip地址
        :return: ip
        """
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(('8.8.8.8', 80))
            ip = s.getsockname()[0]
        except Exception:
            return "unknown"
        return ip

    def load_transmit(self):
        count = 0
        while count <= 5:
            time.sleep(2)
            syslog.syslog(syslog.LOG_DEBUG, "reload transmit, count = {}".format(count))
            try:
                self.log = SystemBus().get("org.log.sys_transmit", "/org/log/sys_transmit")
                break
            except Exception:
                count = count + 1
                syslog.syslog(syslog.LOG_DEBUG, "log service is error")

    def log_transmit(self, name, message: str):
        time = self.get_host_time
        hostname = self.get_host_name
        ip = self.get_host_ip
        log_message = "'time':'{}','hostname':'{}','ip':'{}','name':'{}','lv':'info','message':'{}'".format(time, hostname, ip, name, message)
        syslog.syslog(syslog.LOG_DEBUG, "log_messgae = {}".format(log_message))
        try:
            ret = self.log.log_transmit('{' + log_message + '}')
            syslog.syslog(syslog.LOG_DEBUG, "sent log ret = {}".format(ret))
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "send log message failed")

    def getProcess(self):
        # return self.officeExecList
        path = []
        try:
            f = open(self.keyProcessListPath, "r")
            path = f.read().splitlines()
            syslog.syslog(syslog.LOG_DEBUG, "get key process path = {}".format(path))
            f.close()
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "{} does not exist".format(self.keyProcessListPath))
        return path

    def saveProcessList(self, list):
        f = open(self.keyProcessListPath, "w")
        for path in self.officeExecList:
            f.writelines(path + "\n")
            syslog.syslog(syslog.LOG_DEBUG, "save key process path = {}".format(path))
        f.close()

    def readProcessList(self):
        process_list = []
        try:
            f = open(self.keyProcessListPath, "r")
            process_list = f.read().splitlines()
            syslog.syslog(syslog.LOG_DEBUG, "read process list = {}".format(process_list))
            f.close()
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "{} does not exist".format(self.keyProcessListPath))
        return process_list

    def setProcess(self, list):
        tmplist =  self.officeExecList
        try:
            self.officeExecList = json.loads(list)
            syslog.syslog(syslog.LOG_DEBUG, "setProcess list = {}".format(self.officeExecList))
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "set process list is invalid")
            return -1

        controller = KeyProcessControl()
        ret = controller.kysec_process_enable()
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "enable kysec process protect failed")
            return ret

        for path in tmplist:
            ret = controller.kysec_ppro_delete(path)
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "delete kysec process {} failed.".format(path))
        
        for path in self.officeExecList:
            ret = controller.kysec_ppro_add(path)
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "add {} failed".format(path))
        self.saveProcessList(self.officeExecList)
        return ret

    def getCommandbyPid(self, pid) -> str:
        for pi in self.processList:
            syslog.syslog(syslog.LOG_DEBUG, "------------ pi.pid =  {} command = {}".format(pi.pid, pi.command))
            if(pi.pid == pid):
                return pi.command
        return ""

    @staticmethod
    def logMonitor(self):
        syslog.syslog(syslog.LOG_DEBUG, "logMonitor start")
        with open(self.auditlogPath) as f:
            f.seek(0,2)
            file_size = os.path.getsize(self.auditlogPath)
            while True:
                schedule.run_pending()
                new_file_size = os.path.getsize(self.auditlogPath)
                if new_file_size < file_size:
                    syslog.syslog(syslog.LOG_DEBUG, "reload audit.log")
                    f.seek(0,2)
                data = f.readlines()
                
                for line in data: 
                    if line == '\n':    #空行不处理
                        continue
                    self.parseLog(line)

                file_size = new_file_size
                time.sleep(1)
                # 处理文件截断问题

    def sendOfficeInfo(self, pi:processInfo):
        message = "status:" + pi.status + ", comm:" + pi.apath
        self.log_transmit("processmonitor", message)

    def dailySend(self):
        syslog.syslog(syslog.LOG_DEBUG, "dailySend: get officeExecList = {}".format(self.officeExecList))
        #零点发送还在运行的关键应用、重置processList
        tmplist = []
        for pi in self.processList:
                if os.path.exists("/proc/" + pi.pid):
                    tmplist.append(pi)
                    if(pi.apath in self.officeExecList):
                        syslog.syslog(syslog.LOG_DEBUG, "dailySend: get run pid = {}, status = {}, comm = {}, apath = {}".format(pi.pid, pi.status, pi.command, pi.apath))
                        self.sendOfficeInfo(pi)
        self.processList = tmplist

    def parseLog(self, line):
        if "{ run }" in line:
            syslog.syslog(syslog.LOG_DEBUG, "get run log: {}".format(line))
            p = 'type=.* msg=audit\(([0-9]{10}).*\).*?{ run } for msg="pid=(.*) comm=.*" loginuid=(.*) session=.*'
            info = re.findall(p, line)
            syslog.syslog(syslog.LOG_DEBUG, "get run pid: {}".format(info[0][1]))
            syslog.syslog(syslog.LOG_DEBUG, "self.officeExecList = {}".format(self.officeExecList))

            pi = processInfo(info[0][1])
            if pi.apath in self.officeExecList:
                self.sendOfficeInfo(pi)
            self.processList.append(pi)

        if "{ exit }" in line:
            syslog.syslog(syslog.LOG_DEBUG, "get exit log: {}".format(line))
            p = 'type=.* msg=audit\(([0-9]{10}).*\).*?{ exit } for msg="pid=(.*) comm=.*'
            info = re.findall(p, line)
            syslog.syslog(syslog.LOG_DEBUG, "get exit pid: {}".format(info[0][1]))
            syslog.syslog(syslog.LOG_DEBUG, "self.officeExecList = {}".format(self.officeExecList))

            for pi in self.processList:
                syslog.syslog(syslog.LOG_DEBUG, "processList: pid = {}, status = {}, comm = {}, apath = {}".format(pi.pid, pi.status, pi.command, pi.apath))
                if pi.pid != info[0][1]:
                    continue

                pi.status = "exit"
                if pi.apath in self.officeExecList:
                    self.sendOfficeInfo(pi)

if __name__ == "__main__":
    loop = GLib.MainLoop()
    bus = SystemBus()
    bus.publish("com.kylin.processMonitor", processMonitor())
    loop.run()
